"""
Copyright 2022 (C) Friedrich Miescher Institute for Biomedical Research and
University of Zurich

Original authors:
Tommaso Comparin <tommaso.comparin@exact-lab.it>
Marco Franzon <marco.franzon@exact-lab.it>

This file is part of Fractal and was originally developed by eXact lab S.r.l.
<exact-lab.it> under contract with Liberali Lab from the Friedrich Miescher
Institute for Biomedical Research and Pelkmans Lab from the University of
Zurich.
"""
import json
import logging
import warnings
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Optional

import anndata as ad
import dask
import dask.array as da
import numpy as np
from skimage.io import imread

from fractal_tasks_core.lib_pyramid_creation import write_pyramid
from fractal_tasks_core.lib_regions_of_interest import (
    convert_ROI_table_to_indices,
)
from fractal_tasks_core.lib_zattrs_utils import extract_zyx_pixel_sizes


def correct(
    img,
    illum_img=None,
    background=110,
):
    """
    Corrects single Z level input image using an illumination profile

    The illumination profile image can be uint8 or uint16.
    It needs to follow the illumination profile. e.g. bright in the
    center of the image, dim outside

    :param img: input image to be corrected, either uint8 or uint16
    :type img: np.array
    :param illum_img: correction matrix
    :type illum_img: np.array
    :param background: value for background subtraction (optional, default 110)
    :type background: int

    """

    logging.info(f"Start correct function on image of shape {img.shape}")

    # Check shapes
    if illum_img.shape != img.shape:
        raise Exception(
            "Error in illumination_correction\n"
            f"img.shape: {img.shape}\n"
            f"illum_img.shape: {illum_img.shape}"
        )

    # Background subtraction
    # FIXME: is there a problem with these changes?
    # devdoc.net/python/dask-2.23.0-doc/delayed-best-practices.html
    # ?highlight=delayed#don-t-mutate-inputs
    img[img <= background] = 0
    img[img > background] = img[img > background] - background

    # Apply the illumination correction
    # (normalized by the max value in the illum_img)
    img_corr = img / (illum_img / np.max(illum_img))

    # Handle edge case: The illumination correction can increase a value
    # beyond the limit of the encoding, e.g. beyond 65535 for 16bit
    # images. This clips values that surpass this limit and triggers
    # a warning
    if np.sum(img_corr > np.iinfo(img.dtype).max) > 0:
        warnings.warn(
            f"The illumination correction created values \
                       beyond the max range of your current image \
                       type. These have been clipped to \
                       {np.iinfo(img.dtype).max}"
        )
        img_corr[img_corr > np.iinfo(img.dtype).max] = np.iinfo(img.dtype).max

    logging.info("End correct function")

    return img_corr.astype(img.dtype)


def illumination_correction(
    *,
    input_paths: Iterable[Path],
    output_path: Path,
    metadata: Optional[Dict[str, Any]] = None,
    component: str = None,
    overwrite: bool = False,
    dict_corr: dict = None,
    background: int = 100,
):

    """
    FIXME

    Example inputs:
    input_paths: [PosixPath('tmp_out/*.zarr')]
    output_path: PosixPath('tmp_out/*.zarr')
    component: myplate.zarr/B/03/0/
    metadata: {...}
    """

    # Read some parameters from metadata
    num_levels = metadata["num_levels"]
    coarsening_xy = metadata["coarsening_xy"]
    chl_list = metadata["channel_list"]
    plate, well = component.split(".zarr/")

    if not overwrite:
        raise NotImplementedError("Only overwrite=True currently supported")

    # Define zarrurl
    if len(input_paths) > 1:
        raise NotImplementedError
    in_path = input_paths[0]
    zarrurl = (in_path.parent.resolve() / component).as_posix() + "/"

    # Sanitize zarr paths
    newzarrurl = zarrurl

    logging.info(
        "Start illumination_correction " f"with {zarrurl=} and {newzarrurl=}"
    )

    # Read FOV ROIs
    FOV_ROI_table = ad.read_zarr(f"{zarrurl}tables/FOV_ROI_table")

    # Read pixel sizes from zattrs file
    full_res_pxl_sizes_zyx = extract_zyx_pixel_sizes(
        zarrurl + ".zattrs", level=0
    )

    # Create list of indices for 3D FOVs spanning the entire Z direction
    list_indices = convert_ROI_table_to_indices(
        FOV_ROI_table,
        level=0,
        coarsening_xy=coarsening_xy,
        full_res_pxl_sizes_zyx=full_res_pxl_sizes_zyx,
    )

    # Extract image size from FOV-ROI indices
    # Note: this works at level=0, where FOVs should all be of the exact same
    #       size (in pixels)
    ref_img_size = None
    for indices in list_indices:
        img_size = (indices[3] - indices[2], indices[5] - indices[4])
        if ref_img_size is None:
            ref_img_size = img_size
        else:
            if img_size != ref_img_size:
                raise Exception(
                    "ERROR: inconsistent image sizes in list_indices"
                )
    img_size_y, img_size_x = img_size[:]

    # Load paths of correction matrices
    root_path_corr = dict_corr.pop("root_path_corr")
    if not root_path_corr.endswith("/"):
        root_path_corr += "/"

    # Assemble dictionary of matrices and check their shapes
    corrections = {}
    for ind_ch, ch in enumerate(chl_list):
        corrections[ch] = imread(root_path_corr + dict_corr[ch])
        if corrections[ch].shape != (img_size_y, img_size_x):
            raise Exception(
                "Error in illumination_correction, "
                "correction matrix has wrong shape."
            )

    # Read number of levels from .zattrs of original zarr file
    with open(zarrurl + ".zattrs", "r") as inputjson:
        zattrs = json.load(inputjson)
    num_levels = len(zattrs["multiscales"][0]["datasets"])

    # Load highest-resolution level from original zarr array
    data_czyx = da.from_zarr(zarrurl + "/0")
    dtype = data_czyx.dtype

    # Check that input array is made of images (in terms of shape/chunks)
    nc, nz, ny, nx = data_czyx.shape
    if (ny % img_size_y != 0) or (nx % img_size_x != 0):
        raise Exception(
            "Error in illumination_correction, "
            f"data_czyx.shape: {data_czyx.shape}"
        )
    chunks_c, chunks_z, chunks_y, chunks_x = data_czyx.chunks
    if len(set(chunks_c)) != 1 or chunks_c[0] != 1:
        raise Exception(
            f"Error in illumination_correction, chunks_c: {chunks_c}"
        )
    if len(set(chunks_z)) != 1 or chunks_z[0] != 1:
        raise Exception(
            f"Error in illumination_correction, chunks_z: {chunks_z}"
        )
    if len(set(chunks_y)) != 1 or chunks_y[0] != img_size_y:
        raise Exception(
            f"Error in illumination_correction, chunks_y: {chunks_y}"
        )
    if len(set(chunks_x)) != 1 or chunks_x[0] != img_size_x:
        raise Exception(
            f"Error in illumination_correction, chunks_x: {chunks_x}"
        )

    # Prepare delayed function
    delayed_correct = dask.delayed(correct)

    # Loop over channels
    data_czyx_new = []
    data_czyx_new = da.empty(
        data_czyx.shape,
        chunks="auto",
        dtype=data_czyx.dtype,
    )
    for ind_ch, ch in enumerate(chl_list):
        # Set correction matrix
        illum_img = corrections[ch]
        # 3D data for multiple FOVs
        # Loop over FOVs
        for indices in list_indices:
            s_z, e_z, s_y, e_y, s_x, e_x = indices[:]
            # 3D single-FOV data
            tmp_zyx = []
            # For each FOV, loop over Z planes
            for ind_z in range(e_z):
                shape = [e_y - s_y, e_x - s_x]
                new_img = delayed_correct(
                    data_czyx[ind_ch, ind_z, s_y:e_y, s_x:e_x],
                    illum_img,
                    background=background,
                )
                tmp_zyx.append(da.from_delayed(new_img, shape, dtype))
            data_czyx_new[ind_ch, s_z:e_z, s_y:e_y, s_x:e_x] = da.stack(
                tmp_zyx, axis=0
            )
    tmp_accumulated_data = data_czyx_new
    accumulated_data = tmp_accumulated_data.rechunk(data_czyx.chunks)

    # Construct resolution pyramid
    write_pyramid(
        accumulated_data,
        newzarrurl=newzarrurl,
        overwrite=overwrite,
        coarsening_xy=coarsening_xy,
        num_levels=num_levels,
        chunk_size_x=img_size_x,
        chunk_size_y=img_size_y,
    )

    logging.info("End illumination_correction")


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser(prog="illumination_correction.py")
    parser.add_argument(
        "-z", "--zarrurl", help="zarr url, at the FOV level", required=True
    )
    parser.add_argument(
        "-o",
        "--overwrite",
        action="store_true",
        help="overwrite original zarr file",
    )
    parser.add_argument(
        "-znew",
        "--newzarrurl",
        help="path of the new zarr file",
    )
    parser.add_argument(
        "--path_dict_corr",
        help="path of JSON file with info on illumination matrices",
    )
    parser.add_argument(
        "-C",
        "--chl_list",
        nargs="+",
        help="list of channel names (e.g. A01_C01)",
    )
    parser.add_argument(
        "-cxy",
        "--coarsening_xy",
        default=2,
        type=int,
        help="coarsening factor along X and Y (optional, defaults to 2)",
    )
    parser.add_argument(
        "-bg",
        "--background",
        default=110,
        type=int,
        help=(
            "threshold for background subtraction"
            " (optional, defaults to 110)"
        ),
    )

    args = parser.parse_args()
    illumination_correction(
        args.zarrurl,
        overwrite=args.overwrite,
        newzarrurl=args.newzarrurl,
        path_dict_corr=args.path_dict_corr,
        chl_list=args.chl_list,
        coarsening_xy=args.coarsening_xy,
        background=args.background,
    )



"""
Copyright 2022 (C) Friedrich Miescher Institute for Biomedical Research and
University of Zurich

Original authors:
Tommaso Comparin <tommaso.comparin@exact-lab.it>
Marco Franzon <marco.franzon@exact-lab.it>

This file is part of Fractal and was originally developed by eXact lab S.r.l.
<exact-lab.it> under contract with Liberali Lab from the Friedrich Miescher
Institute for Biomedical Research and Pelkmans Lab from the University of
Zurich.
"""
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Optional

import dask.array as da
from devtools import debug

from fractal_tasks_core.lib_pyramid_creation import write_pyramid


def maximum_intensity_projection(
    *,
    input_paths: Iterable[Path],
    output_path: Path,
    metadata: Optional[Dict[str, Any]] = None,
    component: str = None,
):

    """
    Perform maximum-intensity projection along Z axis, and store the output in
    a new zarr file.

    Examples
      input_paths[0] = /tmp/out_mip/*.zarr  (Path)
      output_path = /tmp/out_mip/*zarr   (Path)
      metadata = {"num_levels": 2, "coarsening_xy": 2, ...}
      component = plate.zarr/B/03/0     (str)
    """

    # Preliminary checks
    if len(input_paths) > 1:
        raise NotImplementedError

    # Read some parameters from metadata
    num_levels = metadata["num_levels"]
    coarsening_xy = metadata["coarsening_xy"]
    plate, well = component.split(".zarr/")

    zarrurl_old = metadata["replicate_zarr"]["sources"][plate] + "/" + well
    clean_output_path = output_path.parent.resolve()
    zarrurl_new = (clean_output_path / component).as_posix()
    debug(zarrurl_old)
    debug(zarrurl_new)

    # Hard-coded values (by now) of chunk sizes to be passed to rechunk,
    # both at level 0 (before coarsening) and at levels 1,2,.. (after
    # repeated coarsening).
    # Note that balance=True may override these values.
    # FIXME should this be inferred from somewhere?
    chunk_size_x = 2560
    chunk_size_y = 2160

    # Load 0-th level
    data_czyx = da.from_zarr(zarrurl_old + "/0")
    num_channels = data_czyx.shape[0]
    # Loop over channels
    accumulate_chl = []
    for ind_ch in range(num_channels):
        # Perform MIP for each channel of level 0
        mip_yx = da.stack([da.max(data_czyx[ind_ch], axis=0)], axis=0)
        accumulate_chl.append(mip_yx)
    accumulate_chl = da.stack(accumulate_chl, axis=0)

    # Construct resolution pyramid
    write_pyramid(
        accumulate_chl,
        newzarrurl=zarrurl_new,
        overwrite=False,
        coarsening_xy=coarsening_xy,
        num_levels=num_levels,
        chunk_size_x=chunk_size_x,
        chunk_size_y=chunk_size_y,
    )


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser(prog="maximum_intensity_projection.py")
    parser.add_argument(
        "-z", "--zarrurl", help="zarr url, at the FOV level", required=True
    )

    parser.add_argument(
        "-cxy",
        "--coarsening_xy",
        default=2,
        type=int,
        help="coarsening factor along X and Y (optional, defaults to 2)",
    )

    args = parser.parse_args()
    maximum_intensity_projection(args.zarrurl, args.coarsening_xy)



"""
Copyright 2022 (C) Friedrich Miescher Institute for Biomedical Research and
University of Zurich

Original authors:
Joel Lüthi  <joel.luethi@fmi.ch>

This file is part of Fractal and was originally developed by eXact lab S.r.l.
<exact-lab.it> under contract with Liberali Lab from the Friedrich Miescher
Institute for Biomedical Research and Pelkmans Lab from the University of
Zurich.
"""
import warnings

import numpy as np
import pandas as pd
from defusedxml import ElementTree


def parse_yokogawa_metadata(mrf_path, mlf_path):
    """
    Parse Yokogawa CV7000 metadata files and prepare site-level metadata

    :param mrf_path: Full path to MeasurementDetail.mrf metadata file
    :type mrf_path: Union(str, pathlib.Path)
    :param mlf_path: Full path to MeasurementData.mlf metadata file
    :type mlf_path: Union(str, pathlib.Path)
    """
    mrf_frame, mlf_frame, error_count = read_metadata_files(mrf_path, mlf_path)

    per_site_parameters = [
        "x_micrometer",
        "y_micrometer",
        "pixel_size_x",
        "pixel_size_y",
        "x_pixel",
        "y_pixel",
        "bit_depth",
    ]
    grouping_params = ["well_id", "field_id"]
    grouped_sites = mlf_frame.loc[
        :, grouping_params + per_site_parameters
    ].groupby(by=grouping_params)
    check_grouped_sites_consistency(grouped_sites, per_site_parameters)
    site_metadata = grouped_sites.mean()

    # Cast image pixel sizes & bit depth to int
    site_metadata = site_metadata.astype(
        {
            "x_pixel": "int",
            "y_pixel": "int",
            "bit_depth": "int",
        }
    )

    # Absolute Z positions are not saved by the Yokogawa,
    # only relative positions to the autofocus
    site_metadata["z_micrometer"] = 0

    site_metadata = pd.concat(
        [
            site_metadata,
            get_z_steps(mlf_frame),
            get_earliest_time_per_site(mlf_frame),
        ],
        axis=1,
    )

    if error_count > 0:
        print(
            f"Succesfully parsed {len(site_metadata)} sites, could not "
            f"parse {error_count} sites due to errors (see warnings)."
        )
    total_files = len(mlf_frame)
    # TODO: Check whether the total_files correspond to the number of
    # relevant input images in the input folder. Returning it for now
    # Maybe return it here for further checks and produce a warning if it does
    # not match

    return site_metadata, total_files


def read_metadata_files(mrf_path, mlf_path):
    # parsing of mrf & mlf files are based on the
    # yokogawa_image_collection_task v0.5 in drogon, written by Dario Vischi.
    # https://github.com/fmi-basel/job-system-workflows/blob/00bbf34448972d27f258a2c28245dd96180e8229/src/gliberal_workflows/tasks/yokogawa_image_collection_task/versions/version_0_5.py  # noqa
    # Now modified for Fractal use

    mrf_frame = read_mrf_file(mrf_path)
    # TODO: filter_position & filter_wheel_position are parsed, but not
    # processed further. Figure out how to save them as relevant metadata for
    # use e.g. during illumination correction

    mlf_frame, error_count = read_mlf_file(mlf_path, mrf_frame)
    # TODO: Time points are parsed as part of the mlf_frame, but currently not
    # processed further. Once we tackle time-resolved data, parse from here.

    return mrf_frame, mlf_frame, error_count


def read_mrf_file(mrf_path):

    # Prepare mrf dataframe
    mrf_columns = [
        "channel_id",
        "horiz_pixel_dim",
        "vert_pixel_dim",
        "camera_no",
        "bit_depth",
        "horiz_pixels",
        "vert_pixels",
        "filter_wheel_position",
        "filter_position",
        "shading_corr_src",
    ]
    mrf_frame = pd.DataFrame(columns=mrf_columns)

    mrf_xml = ElementTree.parse(mrf_path).getroot()
    # Read mrf file
    ns = {"bts": "http://www.yokogawa.co.jp/BTS/BTSSchema/1.0"}
    for channel in mrf_xml.findall("bts:MeasurementChannel", namespaces=ns):
        mrf_frame.loc[channel.get("{%s}Ch" % ns["bts"])] = [
            channel.get("{%s}Ch" % ns["bts"]),
            float(channel.get("{%s}HorizontalPixelDimension" % ns["bts"])),
            float(channel.get("{%s}VerticalPixelDimension" % ns["bts"])),
            int(channel.get("{%s}CameraNumber" % ns["bts"])),
            int(channel.get("{%s}InputBitDepth" % ns["bts"])),
            int(channel.get("{%s}HorizontalPixels" % ns["bts"])),
            int(channel.get("{%s}VerticalPixels" % ns["bts"])),
            int(channel.get("{%s}FilterWheelPosition" % ns["bts"])),
            int(channel.get("{%s}FilterPosition" % ns["bts"])),
            channel.get("{%s}ShadingCorrectionSource" % ns["bts"]),
        ]

    return mrf_frame


def read_mlf_file(mlf_path, mrf_frame):
    # Docstring TBD
    mlf_xml = ElementTree.parse(mlf_path).getroot()
    ns = {"bts": "http://www.yokogawa.co.jp/BTS/BTSSchema/1.0"}

    # Measure number of lines
    def blocks(fh, size=65536):
        while True:
            block = fh.read(size)
            if not block:
                break
            yield block

    mlf_entries = mlf_xml.findall("bts:MeasurementRecord", namespaces=ns)
    nb_lines = len(mlf_entries)

    # Prepare mlf dataframe
    mlf_columns = [
        "type",
        "well_id",
        "column",
        "row",
        "time_point",
        "field_id",
        "z_index",
        "timeline_id",
        "action_id",
        "action",
        "x_micrometer",
        "y_micrometer",
        "z_micrometer",
        "x_pixel",
        "y_pixel",
        "pixel_size_x",
        "pixel_size_y",
        "bit_depth",
        "width",
        "height",
        "channel_id",
        "camera_no",
        "file_name",
        "time",
    ]
    mlf_frame = pd.DataFrame(columns=mlf_columns, index=range(0, nb_lines))

    mrf_channel_indices = {
        row.channel_id: idx
        for idx, (_, row) in enumerate(mrf_frame.iterrows())
    }

    error_count = 0
    for idx, record in enumerate(mlf_entries):
        rec_type = record.get("{%s}Type" % ns["bts"])

        if rec_type == "ERR":
            warnings.warn(
                "When parsing the yokogawa metadata, "
                f"found an 'ERR' entry at line '{idx+3}' in the MLF "
                f"meta file! The error was: '{record.text}'."
                "The entry is skipped."
            )
            error_count += 1
            continue
        elif rec_type != "IMG":
            warnings.warn(
                "When parsing the yokogawa metadata, "
                f"Found unexpected '{rec_type}' entry at line '{idx+3}'"
                "in the MLF meta file! The entry is skipped."
            )
            error_count += 1
            continue

        channel_id = record.get("{%s}Ch" % ns["bts"])
        x_micrometer = float(record.get("{%s}X" % ns["bts"]))
        # we mirror the y coordinate to fit with the field layout
        y_micrometer = -float(record.get("{%s}Y" % ns["bts"]))
        z_micrometer = float(record.get("{%s}Z" % ns["bts"]))

        well_row_id = record.get("{%s}Row" % ns["bts"])
        well_col_id = record.get("{%s}Column" % ns["bts"])
        well_id = chr(64 + int(well_row_id)) + str(well_col_id).zfill(2)
        # Convert all times to UTC time zone to avoid later timezone handling
        time = pd.to_datetime(record.get("{%s}Time" % ns["bts"]), utc=True)

        bit_depth = np.nan
        width = np.nan
        height = np.nan
        camera_no = np.nan
        pixel_size_x = np.nan
        pixel_size_y = np.nan
        if rec_type == "IMG":
            mrf_idx = mrf_channel_indices[channel_id]
            pixel_size_x = mrf_frame.iat[mrf_idx, 1]
            pixel_size_y = mrf_frame.iat[mrf_idx, 2]
            bit_depth = int(mrf_frame.iat[mrf_idx, 4])
            width = int(mrf_frame.iat[mrf_idx, 5])
            height = int(mrf_frame.iat[mrf_idx, 6])
            camera_no = int(mrf_frame.iat[mrf_idx, 3])

        # we use iat[] here for (significant) performance reasons
        mlf_frame.iat[idx, 0] = rec_type
        mlf_frame.iat[idx, 1] = well_id
        mlf_frame.iat[idx, 2] = int(well_col_id)
        mlf_frame.iat[idx, 3] = int(well_row_id)
        mlf_frame.iat[idx, 4] = int(record.get("{%s}TimePoint" % ns["bts"]))
        mlf_frame.iat[idx, 5] = int(record.get("{%s}FieldIndex" % ns["bts"]))
        mlf_frame.iat[idx, 6] = int(record.get("{%s}ZIndex" % ns["bts"]))
        mlf_frame.iat[idx, 7] = int(
            record.get("{%s}TimelineIndex" % ns["bts"])
        )
        mlf_frame.iat[idx, 8] = int(record.get("{%s}ActionIndex" % ns["bts"]))
        mlf_frame.iat[idx, 9] = record.get("{%s}Action" % ns["bts"])
        mlf_frame.iat[idx, 10] = x_micrometer
        mlf_frame.iat[idx, 11] = y_micrometer
        mlf_frame.iat[idx, 12] = z_micrometer
        mlf_frame.iat[idx, 13] = width
        mlf_frame.iat[idx, 14] = height
        mlf_frame.iat[idx, 15] = pixel_size_x
        mlf_frame.iat[idx, 16] = pixel_size_y

        mlf_frame.iat[idx, 17] = bit_depth
        mlf_frame.iat[idx, 18] = width
        mlf_frame.iat[idx, 19] = height
        mlf_frame.iat[idx, 20] = channel_id
        mlf_frame.iat[idx, 21] = camera_no
        mlf_frame.iat[idx, 22] = record.text  # file_name
        mlf_frame.iat[idx, 23] = time  # file_name

    mlf_frame = mlf_frame.dropna(thresh=(len(mlf_frame.columns)))
    return mlf_frame, error_count


def calculate_steps(site_series: pd.Series):
    # site_series is the z_micrometer series for a given site of a given
    # channel. This function calculates the step size in Z

    # First diff is always NaN because there is nothing to compare it to
    steps = site_series.diff().dropna()
    if not steps.std().sum() == 0.0:
        raise Exception(
            "When parsing the Yokogawa mlf file, some sites "
            "had varying step size in Z. "
            "That is not supported for the OME-Zarr parsing"
        )
    return steps.mean()


def get_z_steps(mlf_frame):
    # Process mlf_frame to extract Z information (pixel size & steps).
    # Run checks on consistencies & return site-based z step dataframe
    # Group by well, field & channel
    grouped_sites_z = (
        mlf_frame.loc[
            :,
            ["well_id", "field_id", "action_id", "channel_id", "z_micrometer"],
        ]
        .set_index(["well_id", "field_id", "action_id", "channel_id"])
        .groupby(level=[0, 1, 2, 3])
    )

    # If there is only 1 Z step, set the Z spacing to the count of planes => 1
    if grouped_sites_z.count()["z_micrometer"].max() == 1:
        z_data = grouped_sites_z.count().groupby(["well_id", "field_id"])
    else:
        # Group the whole site (combine channels), because Z steps need to be
        # consistent between channels for OME-Zarr.
        z_data = grouped_sites_z.apply(calculate_steps).groupby(
            ["well_id", "field_id"]
        )
    if not z_data.std().sum().sum() == 0.0:
        raise Exception(
            "When parsing the Yokogawa mlf file, channels had "
            "varying step size in Z. "
            "That is not supported for the OME-Zarr parsing"
        )

    # Ensure that channels have the same number of z planes and
    # reduce it to one value.
    # Only check if there is more than one channel available
    if any(
        grouped_sites_z.count().groupby(["well_id", "field_id"]).count() > 1
    ):
        if any(
            grouped_sites_z.count()
            .groupby(["well_id", "field_id"])
            .std()
            .sum()
            != 0
        ):
            raise Exception(
                "When parsing the Yokogawa mlf file, channels had "
                "varying number of z planes."
                "That is not supported for the OME-Zarr parsing"
            )
    z_steps = (
        grouped_sites_z.count()
        .groupby(["well_id", "field_id"])
        .mean()
        .astype(int)
    )

    # Combine the two dataframes
    z_frame = pd.concat([z_data.mean(), z_steps], axis=1)
    z_frame.columns = ["pixel_size_z", "z_pixel"]
    return z_frame


def check_grouped_sites_consistency(grouped_sites, per_site_parameters):
    # Check that stage positions don't vary within a given site
    # Same for pixel sizes
    # Only relevant when a site has multiple entries
    if grouped_sites.count().min().min() > 1:
        if not grouped_sites.std().sum().sum() == 0.0:
            raise Exception(
                ""
                "When parsing the Yokogawa MeasurementData.mlf file, "
                f"some of the parameters {per_site_parameters} varied within "
                "the site. That is not supported for the OME-Zarr parsing"
            )


def get_earliest_time_per_site(mlf_frame) -> pd.DataFrame:
    # Get the time information per site
    # Because a site will contain time information for each plane
    # of each channel, we just return the earliest time infromation
    # per site.
    return mlf_frame.groupby(["well_id", "field_id"]).min()["time"]



"""
Copyright 2022 (C) Friedrich Miescher Institute for Biomedical Research and
University of Zurich

Original authors:
Tommaso Comparin <tommaso.comparin@exact-lab.it>
Marco Franzon <marco.franzon@exact-lab.it>

This file is part of Fractal and was originally developed by eXact lab S.r.l.
<exact-lab.it> under contract with Liberali Lab from the Friedrich Miescher
Institute for Biomedical Research and Pelkmans Lab from the University of
Zurich.
"""
import os
import re
from glob import glob
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Optional

import dask.array as da
from anndata import read_zarr
from skimage.io import imread

from fractal_tasks_core.lib_pyramid_creation import write_pyramid
from fractal_tasks_core.lib_regions_of_interest import (
    convert_ROI_table_to_indices,
)
from fractal_tasks_core.lib_zattrs_utils import extract_zyx_pixel_sizes


def sort_fun(s):
    """
    sort_fun takes a string (filename of a yokogawa images),
    extract site and z-index metadata and returns them as a list.

    :param s: filename
    :type s: str
    """

    site = re.findall(r"F(.*)L", s)[0]
    zind = re.findall(r"Z(.*)C", s)[0]
    return [site, zind]


def yokogawa_to_zarr(
    *,
    input_paths: Iterable[Path],
    output_path: Path,
    delete_input=False,
    metadata: Optional[Dict[str, Any]] = None,
    component: str = None,
):
    """
    Convert Yokogawa output (png, tif) to zarr file

    Example arguments:
      input_paths[0] = /tmp/output/*.zarr  (Path)
      output_path = /tmp/output/*.zarr      (Path)
      metadata = {"channel_list": [...], "num_levels": ..., }
      component = plate.zarr/B/03/0/
    """

    # Preliminary checks
    if len(input_paths) > 1:
        raise NotImplementedError

    chl_list = metadata["channel_list"]
    original_path_list = metadata["original_paths"]
    in_path = Path(original_path_list[0]).parent
    ext = Path(original_path_list[0]).name
    num_levels = metadata["num_levels"]
    coarsening_xy = metadata["coarsening_xy"]

    # Hard-coded values (by now) of chunk sizes to be passed to rechunk,
    # both at level 0 (before coarsening) and at levels 1,2,.. (after
    # repeated coarsening).
    # Note that balance=True may override these values.
    chunk_size_x = 2560
    chunk_size_y = 2160

    # Define well
    component_split = component.split("/")
    well_row = component_split[1]
    well_column = component_split[2]

    well_ID = well_row + well_column

    # delayed_imread = delayed(imread)

    print(f"Channels: {chl_list}")

    list_channels = []
    for chl in chl_list:
        A, C = chl.split("_")

        glob_path = f"{in_path}/*_{well_ID}_*{A}*{C}{ext}"
        print(f"glob path: {glob_path}")
        filenames = sorted(glob(glob_path), key=sort_fun)
        if len(filenames) == 0:
            raise Exception(
                "Error in yokogawa_to_zarr: len(filenames)=0.\n"
                f"  in_path: {in_path}\n"
                f"  ext: {ext}\n"
                f"  well_ID: {well_ID}\n"
                f"  channel: {chl},\n"
                f"  glob_path: {glob_path}"
            )

        sample = imread(filenames[0])

        zarrurl = input_paths[0].parent.as_posix() + f"/{component}"
        adata = read_zarr(f"{zarrurl}/tables/FOV_ROI_table")
        pxl_size = extract_zyx_pixel_sizes(f"{zarrurl}/.zattrs")
        fov_position = convert_ROI_table_to_indices(
            adata, full_res_pxl_sizes_zyx=pxl_size
        )

        max_x = max(roi[5] for roi in fov_position)
        max_y = max(roi[3] for roi in fov_position)
        max_z = max(roi[1] for roi in fov_position)

        img_position = []
        for fov in fov_position:
            for z in range(fov[1]):
                img = [z, z + 1, fov[2], fov[3], fov[4], fov[5]]
                img_position.append(img)

        canvas = da.zeros(
            (max_z, max_y, max_x),
            dtype=sample.dtype,
            chunks=(1, chunk_size_y, chunk_size_x),
        )

        for indexes, image_file in zip(*(img_position, filenames)):
            canvas[
                indexes[0] : indexes[1],  # noqa: 203
                indexes[2] : indexes[3],  # noqa: 203
                indexes[4] : indexes[5],  # noqa: 203
            ] = imread(image_file)

        list_channels.append(canvas)
    data_czyx = da.stack(list_channels, axis=0)

    if delete_input:
        for f in filenames:
            try:
                os.remove(f)
            except OSError as e:
                print("Error: %s : %s" % (f, e.strerror))

    # Construct resolution pyramid
    write_pyramid(
        data_czyx,
        newzarrurl=output_path.parent.as_posix() + f"/{component}",
        overwrite=False,
        coarsening_xy=coarsening_xy,
        num_levels=num_levels,
        chunk_size_x=chunk_size_x,
        chunk_size_y=chunk_size_y,
    )


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser(prog="Yokogawa_to_zarr")

    parser.add_argument(
        "-i", "--in_path", help="directory containing the input files"
    )

    parser.add_argument(
        "-z",
        "--zarrurl",
        help="structure of the zarr folder",
    )

    parser.add_argument(
        "-e",
        "--ext",
        help="source images extension",
    )

    parser.add_argument(
        "-C",
        "--chl_list",
        nargs="+",
        help="list of channel names (e.g. A01_C01)",
    )

    parser.add_argument(
        "-nl",
        "--num_levels",
        type=int,
        help="number of levels in the Zarr pyramid",
    )

    parser.add_argument(
        "-cxy",
        "--coarsening_xy",
        default=2,
        type=int,
        help="coarsening factor along X and Y (optional, defaults to 2)",
    )

    parser.add_argument(
        "-d",
        "--delete_input",
        action="store_true",
        help="Delete input files",
    )

    args = parser.parse_args()

    yokogawa_to_zarr(
        args.zarrurl,
        in_path=args.in_path,
        ext=args.ext,
        chl_list=args.chl_list,
        num_levels=args.num_levels,
        coarsening_xy=args.coarsening_xy,
        delete_input=args.delete_input,
    )



"""
Copyright 2022 (C) Friedrich Miescher Institute for Biomedical Research and
University of Zurich

Original authors:
Tommaso Comparin <tommaso.comparin@exact-lab.it>
Marco Franzon <marco.franzon@exact-lab.it>

This file is part of Fractal and was originally developed by eXact lab S.r.l.
<exact-lab.it> under contract with Liberali Lab from the Friedrich Miescher
Institute for Biomedical Research and Pelkmans Lab from the University of
Zurich.
"""
import shutil

import dask.array as da


def to_zarr_custom(newzarrurl=None, array=None, component="", overwrite=False):

    """

    Custom workaround for dask issue
    (https://github.com/dask/dask/issues/5942), where a dask array loaded with
    from_zarr cannot be written with to_zarr(..., overwrite=True).

    :param newzarrurl: output zarr file
    :type newzarrurl: str
    :param array: dask array to be stored
    :type array: dask array
    :param component: target subfolder of the zarr file (optional, default "")
    :type component: str
    :param overwrite: overwrite existing data (optional, default False)
    :type overwrite: boolean

    """

    # Check required arguments
    if newzarrurl is None:
        raise Exception("ERROR: Missing newzarrurl arg in to_zarr_custom")
    if array is None:
        raise Exception("ERROR: Missing array arg in to_zarr_custom")

    # Sanitize arguments
    if component.endswith("/"):
        component = component[:-1]
    if not newzarrurl.endswith("/"):
        newzarrurl += "/"

    if overwrite:
        tmp_suffix = "_TEMPORARY_ZARR_ARRAY"
        array.to_zarr(
            newzarrurl,
            component=component + tmp_suffix,
            dimension_separator="/",
            compute=True,
            write_empty_chunks=False,
        )
        shutil.rmtree(newzarrurl + component)
        shutil.move(
            newzarrurl + component + tmp_suffix, newzarrurl + component
        )
        output = da.from_zarr(newzarrurl, component=component)
    else:
        output = array.to_zarr(
            newzarrurl,
            component=component,
            dimension_separator="/",
            compute=True,
            return_stored=True,
            write_empty_chunks=False,
        )

    return output



"""
Copyright 2022 (C) Friedrich Miescher Institute for Biomedical Research and
University of Zurich

Original authors:
Tommaso Comparin <tommaso.comparin@exact-lab.it>
Marco Franzon <marco.franzon@exact-lab.it>

This file is part of Fractal and was originally developed by eXact lab S.r.l.
<exact-lab.it> under contract with Liberali Lab from the Friedrich Miescher
Institute for Biomedical Research and Pelkmans Lab from the University of
Zurich.
"""
import json
import shutil
import time
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Optional

import anndata as ad
import dask
import dask.array as da
import numpy as np
import zarr
from cellpose import core
from cellpose import models
from devtools import debug

from fractal_tasks_core.lib_pyramid_creation import write_pyramid
from fractal_tasks_core.lib_regions_of_interest import convert_ROI_table_to_indices
from fractal_tasks_core.lib_zattrs_utils import extract_zyx_pixel_sizes
from fractal_tasks_core.lib_zattrs_utils import rescale_datasets


def segment_FOV(
    column,
    block_info=None,
    model=None,
    do_3D=True,
    anisotropy=None,
    diameter=40.0,
    cellprob_threshold=0.0,
    flow_threshold=0.4,
    label_dtype=None,
    logfile="LOG_image_labeling",
):

    # Write some debugging info
    with open(logfile, "a") as out:
        out.write(
            f"[segment_FOV] START Cellpose |"
            f" column: {type(column)}, {column.shape} |"
            f" do_3D: {do_3D} |"
            f" model.diam_mean: {model.diam_mean} |"
            f" diameter: {diameter} |"
            f" flow threshold: {flow_threshold}\n"
        )

    # Actual labeling
    t0 = time.perf_counter()
    mask, flows, styles, diams = model.eval(
        column,
        channels=[0, 0],
        do_3D=do_3D,
        net_avg=False,
        augment=False,
        diameter=diameter,
        anisotropy=anisotropy,
        cellprob_threshold=cellprob_threshold,
        flow_threshold=flow_threshold,
    )
    if not do_3D:
        mask = np.expand_dims(mask, axis=0)
    t1 = time.perf_counter()

    # Write some debugging info
    with open(logfile, "a") as out:
        out.write(
            f"[segment_FOV] END   Cellpose |"
            f" Elapsed: {t1-t0:.4f} seconds |"
            f" mask shape: {mask.shape},"
            f" mask dtype: {mask.dtype} (before recast to {label_dtype}),"
            f" max(mask): {np.max(mask)} |"
            f" model.diam_mean: {model.diam_mean} |"
            f" diameter: {diameter} |"
            f" flow threshold: {flow_threshold}\n"
        )

    return mask.astype(label_dtype)


def image_labeling(
    *,
    input_paths: Iterable[Path],
    output_path: Path,
    metadata: Optional[Dict[str, Any]] = None,
    component: str = None,
    labeling_channel: str = None,
    labeling_level: int = 1,
    num_threads: int = 1,
    relabeling: bool = True,
    anisotropy: float = None,
    diameter_level0: float = 80.0,
    cellprob_threshold: float = 0.0,
    flow_threshold: float = 0.4,
    model_type: str = "nuclei",
):

    """
    Example inputs:
      input_paths: PosixPath('tmp_out_mip/*.zarr')
      output_path: PosixPath('tmp_out_mip/*.zarr')
      component: myplate.zarr/B/03/0/
      metadata: {...}

    """

    # Preliminary check
    if len(input_paths) > 1:
        raise NotImplementedError
    in_path = input_paths[0]

    # Read some parameters from metadata
    num_levels = metadata["num_levels"]
    coarsening_xy = metadata["coarsening_xy"]
    chl_list = metadata["channel_list"]
    plate, well = component.split(".zarr/")

    # Find well ID
    well_ID = well.replace("/", "_")[:-1]
    logfile = f"LOG_image_labeling_{well_ID}"
    debug(well_ID)

    # Find channel index
    if labeling_channel not in chl_list:
        raise Exception(f"ERROR: {labeling_channel} not in {chl_list}")
    ind_channel = chl_list.index(labeling_channel)

    # Set labels dtype
    label_dtype = np.uint32

    zarrurl = (in_path.parent.resolve() / component).as_posix() + "/"
    debug(zarrurl)

    # Load ZYX data
    data_zyx = da.from_zarr(f"{zarrurl}{labeling_level}")[ind_channel]
    debug(data_zyx.shape)

    # Read FOV ROIs
    FOV_ROI_table = ad.read_zarr(f"{zarrurl}tables/FOV_ROI_table")

    # Read pixel sizes from zattrs file
    full_res_pxl_sizes_zyx = extract_zyx_pixel_sizes(
        zarrurl + ".zattrs", level=0
    )

    # Create list of indices for 3D FOVs spanning the entire Z direction
    list_indices = convert_ROI_table_to_indices(
        FOV_ROI_table,
        level=labeling_level,
        coarsening_xy=coarsening_xy,
        full_res_pxl_sizes_zyx=full_res_pxl_sizes_zyx,
    )

    # Extract image size from FOV-ROI indices
    # Note: this works at level=0, where FOVs should all be of the exact same
    #       size (in pixels)
    list_indices_level0 = convert_ROI_table_to_indices(
        FOV_ROI_table,
        level=0,
        full_res_pxl_sizes_zyx=full_res_pxl_sizes_zyx,
    )

    ref_img_size = None
    for indices in list_indices_level0:
        img_size = (indices[3] - indices[2], indices[5] - indices[4])
        if ref_img_size is None:
            ref_img_size = img_size
        else:
            if img_size != ref_img_size:
                raise Exception(
                    "ERROR: inconsistent image sizes in list_indices",
                    list_indices,
                )
    img_size_y, img_size_x = img_size[:]

    # Select 2D/3D behavior and set some parameters
    do_3D = data_zyx.shape[0] > 1
    if do_3D:
        if anisotropy is None:
            # Read pixel sizes from zattrs file
            pxl_zyx = extract_zyx_pixel_sizes(
                zarrurl + ".zattrs", level=labeling_level
            )
            pixel_size_z, pixel_size_y, pixel_size_x = pxl_zyx[:]
            debug(pxl_zyx)
            if not np.allclose(pixel_size_x, pixel_size_y):
                raise Exception(
                    "ERROR: XY anisotropy detected\n"
                    f"pixel_size_x={pixel_size_x}\n"
                    f"pixel_size_y={pixel_size_y}"
                )
            anisotropy = pixel_size_z / pixel_size_x

    # Check model_type
    if model_type not in ["nuclei", "cyto2", "cyto"]:
        raise Exception(f"ERROR model_type={model_type} is not allowed.")

    # Load zattrs file
    zattrs_file = f"{zarrurl}.zattrs"
    with open(zattrs_file, "r") as jsonfile:
        zattrs = json.load(jsonfile)

    # Preliminary checks on multiscales
    multiscales = zattrs["multiscales"]
    if len(multiscales) > 1:
        raise NotImplementedError(
            f"Found {len(multiscales)} multiscales, "
            "but only one is currently supported."
        )
    if "coordinateTransformations" in multiscales[0].keys():
        raise NotImplementedError(
            "global coordinateTransformations at the multiscales "
            "level are not currently supported"
        )

    # Extract num_levels
    num_levels = len(multiscales[0]["datasets"])

    # Extract axes, and remove channel
    new_axes = [ax for ax in multiscales[0]["axes"] if ax["type"] != "channel"]

    # Try to read channel label from OMERO metadata
    try:
        omero_label = zattrs["omero"]["channels"][ind_channel]["label"]
        label_name = f"label_{omero_label}"
    except (KeyError, IndexError):
        label_name = f"label_{ind_channel}"

    # Initialize cellpose
    use_gpu = core.use_gpu()
    model = models.Cellpose(gpu=use_gpu, model_type=model_type)

    # Initialize other things
    with open(logfile, "w") as out:
        out.write(f"Start image_labeling task for {zarrurl}\n")
        out.write(f"relabeling: {relabeling}\n")
        out.write(f"do_3D: {do_3D}\n")
        out.write(f"labeling_level: {labeling_level}\n")
        out.write(f"model_type: {model_type}\n")
        out.write(f"anisotropy: {anisotropy}\n")
        out.write(f"num_threads: {num_threads}\n")
        out.write("Total well shape/chunks:\n")
        out.write(f"{data_zyx.shape}\n")
        out.write(f"{data_zyx.chunks}\n\n")

    # Prepare delayed function
    delayed_segment_FOV = dask.delayed(segment_FOV)

    # Prepare empty mask with correct chunks
    mask = da.empty(
        data_zyx.shape,
        dtype=label_dtype,
        chunks=(1, img_size_y, img_size_x),
    )

    # Map labeling function onto all FOV ROIs
    for indices in list_indices:
        s_z, e_z, s_y, e_y, s_x, e_x = indices[:]
        shape = [e_z - s_z, e_y - s_y, e_x - s_x]
        if min(shape) == 0:
            raise Exception(f"ERROR: ROI indices lead to shape {shape}")
        FOV_mask = delayed_segment_FOV(
            data_zyx[s_z:e_z, s_y:e_y, s_x:e_x],
            model=model,
            do_3D=do_3D,
            anisotropy=anisotropy,
            label_dtype=label_dtype,
            diameter=diameter_level0 / coarsening_xy**labeling_level,
            cellprob_threshold=cellprob_threshold,
            flow_threshold=flow_threshold,
            logfile=logfile,
        )
        mask[s_z:e_z, s_y:e_y, s_x:e_x] = da.from_delayed(
            FOV_mask, shape, label_dtype
        )

    with open(logfile, "a") as out:
        out.write(
            f"mask will have shape {mask.shape} "
            f"and chunks {mask.chunks}\n\n"
        )

    # Rescale datasets (only relevant for labeling_level>0)
    new_datasets = rescale_datasets(
        datasets=multiscales[0]["datasets"],
        coarsening_xy=coarsening_xy,
        reference_level=labeling_level,
    )

    # Write zattrs for labels and for specific label
    # FIXME deal with: (1) many channels, (2) overwriting
    labels_group = zarr.group(f"{zarrurl}labels")
    labels_group.attrs["labels"] = [label_name]
    label_group = labels_group.create_group(label_name)
    label_group.attrs["image-label"] = {"version": "0.4"}
    label_group.attrs["multiscales"] = [
        {
            "name": label_name,
            "version": "0.4",
            "axes": new_axes,
            "datasets": new_datasets,
        }
    ]

    if relabeling:

        # Execute all, and write level-0 mask to disk
        with dask.config.set(pool=ThreadPoolExecutor(num_threads)):
            write_pyramid(
                mask,
                newzarrurl=f"{zarrurl}labels/{label_name}/",
                overwrite=False,
                coarsening_xy=coarsening_xy,
                num_levels=1,
                chunk_size_x=img_size_x,
                chunk_size_y=img_size_y,
                aggregation_function=np.max,
            )

        with open(logfile, "a") as out:
            out.write("\nStart relabeling\n")
        t0 = time.perf_counter()

        # Load non-relabeled mask from disk
        mask = da.from_zarr(zarrurl, component=f"labels/{label_name}/{0}")
        newmask = da.empty(
            shape=mask.shape,
            chunks=mask.chunks,
            dtype=label_dtype,
        )

        # Sequential relabeling
        num_labels_tot = 0
        num_labels_column = 0
        for indices in list_indices:
            s_z, e_z, s_y, e_y, s_x, e_x = indices[:]
            shape = [e_z - s_z, e_y - s_y, e_x - s_x]

            # Extract a specific FOV (=column in 3D, =image in 2D)
            column_mask = mask[s_z:e_z, s_y:e_y, s_x:e_x].compute()
            num_labels_column = np.max(column_mask)

            # Apply re-labeling and update total number of labels
            shift = np.zeros_like(column_mask, dtype=label_dtype)
            shift[column_mask > 0] = num_labels_tot
            num_labels_tot += num_labels_column

            with open(logfile, "a") as out:
                out.write(
                    f"FOV ROI {indices}, "
                    f"num_labels_column={num_labels_column}, "
                    f"num_labels_tot={num_labels_tot}\n"
                )

            # Check that total number of labels is under control
            if num_labels_tot > np.iinfo(label_dtype).max:
                raise Exception(
                    "ERROR in re-labeling:\n"
                    f"Reached {num_labels_tot} labels, "
                    f"but dtype={label_dtype}"
                )
            # Re-assign the chunk to the new array
            newmask[s_z:e_z, s_y:e_y, s_x:e_x] = column_mask + shift

        # FIXME: this is ugly
        shutil.rmtree(zarrurl + f"labels/{label_name}/{0}")

        write_pyramid(
            newmask,
            newzarrurl=f"{zarrurl}labels/{label_name}/",
            overwrite=False,
            coarsening_xy=coarsening_xy,
            num_levels=num_levels,
            chunk_size_x=img_size_x,
            chunk_size_y=img_size_y,
            aggregation_function=np.max,
        )

        t1 = time.perf_counter()
        with open(logfile, "a") as out:
            out.write(f"End relabeling, elapsed: {t1-t0} s\n")

    else:

        # Construct resolution pyramid
        with dask.config.set(pool=ThreadPoolExecutor(num_threads)):
            write_pyramid(
                mask,
                newzarrurl=f"{zarrurl}labels/{label_name}/",
                overwrite=False,
                coarsening_xy=coarsening_xy,
                num_levels=num_levels,
                chunk_size_x=img_size_x,
                chunk_size_y=img_size_y,
                aggregation_function=np.max,
            )

        with open(logfile, "a") as out:
            out.write("\nSkip relabeling\n")

    return {}


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser(prog="image_labeling.py")
    parser.add_argument(
        "-z", "--zarrurl", help="zarr url, at the FOV level", required=True
    )
    parser.add_argument(
        "-C",
        "--chl_list",
        nargs="+",
        help="list of channel names (e.g. A01_C01)",
    )
    parser.add_argument(
        "-cxy",
        "--coarsening_xy",
        default=2,
        type=int,
        help="coarsening factor along X and Y (optional, defaults to 2)",
    )
    parser.add_argument(
        "-lc",
        "--labeling_channel",
        help="name of channel for labeling (e.g. A01_C01)",
    )
    parser.add_argument(
        "--num_threads",
        default=1,
        type=int,
        help="TBD",
    )
    parser.add_argument(
        "-ll",
        "--labeling_level",
        default=0,
        type=int,
        help="TBD",
    )

    args = parser.parse_args()
    image_labeling(
        args.zarrurl,
        coarsening_xy=args.coarsening_xy,
        chl_list=args.chl_list,
        labeling_channel=args.labeling_channel,
        num_threads=args.num_threads,
        labeling_level=args.labeling_level,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fractal_tasks_core']

package_data = \
{'': ['*']}

install_requires = \
['PyQt5-Qt5>=5.15.2,<6.0.0',
 'PyQt5-sip>=12.11.0,<13.0.0',
 'PyQt5>=5.15.7,<6.0.0',
 'anndata>=0.8.0,<0.9.0',
 'cellpose==2',
 'dask>=2022.6,<2022.8',
 'defusedxml>=0.7.1,<0.8.0',
 'graphviz>=0.20,<0.21',
 'imagecodecs>=2022.2.22,<2023.0.0',
 'llvmlite>=0.39.1,<0.40.0',
 'napari-skimage-regionprops>=0.5.3,<0.6.0',
 'napari-workflows>=0.2.3,<0.3.0',
 'ome-zarr>=0.5.1,<0.6.0',
 'pandas>=1.2.0,<2.0.0',
 'scikit-image>=0.19.3,<0.20.0']

setup_kwargs = {
    'name': 'fractal-tasks-core',
    'version': '0.1.3',
    'description': '',
    'long_description': '# fractal-tasks-core\nMain tasks for the Fractal analytics platform\n',
    'author': 'Jacopo Nespolo',
    'author_email': 'jacopo.nespolo@exact-lab.it',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/fractal-analytics-platform/fractal-tasks-core',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
